/* Tokens 89 Command-Line Converter - tokenizer/detokenizer for the TI-89/92+/V200
   Copyright (C) 2003 Kevin Kofler <kevin.kofler@chello.at>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   In addition, as a special exception, Kevin Kofler gives permission
   to link the code of this program with the Microsoft Visual Basic
   runtime libraries required by the Tokens89 OCX, and distribute
   linked combinations including the two. You must obey the GNU
   General Public License in all respects for all of the code used
   other than the Microsoft Visual Basic runtime libraries. If you
   modify this file, you may extend this exception to your version of
   the file, but you are not obligated to do so. If you do not wish to
   do so, delete this exception statement from your version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  */

#define INITGUID
#include <stdio.h>
#include <comcat.h>
#include <winnls.h>

int main(int argc, const char **argv)
{
  if (argc==2 && (!strcmp(argv[1],"/?") || !strcmp(argv[1],"-?")
                  || !strcasecmp(argv[1],"/h") || !strcasecmp(argv[1],"-h")
                  || !strcmp(argv[1],"--help"))) {
    puts("Tokens89 command-line converter\n"
         "Usage: tokens89 infile outfile\n"
         "If \"infile\" is a program or function in TI-GraphLink binary format,"
         " it will be\nconverted to the TI-GraphLink 7-bit ASCII interchange "
         "format.\nIf \"infile\" is a program or function in TI-GraphLink 7-bit"
         " ASCII interchange\nformat, it will be converted to the TI-GraphLink "
         "binary format in tokenized\nform.");
    return 0;
  }

  if (argc<3) {puts("Not enough arguments.");return 1;}
  if (argc>3) {puts("Too many arguments.");return 1;}
  
  FILE *f=fopen(argv[1],"rb");
  if (!f) {puts("Error opening file.");return 2;}
  char buffer[8];
  if (fread(buffer,1,8,f)!=8) goto invalid;
  if (!memcmp(buffer,"**TI89**",8) || !memcmp(buffer,"**TI92P*",8))
    goto detoken;
  if (fseek(f,0,SEEK_END)) goto invalid;
  int l=ftell(f);
  if (fseek(f,0,SEEK_SET)) goto invalid;
  if (l==EOF) goto invalid;
  {
    char filebuffer[l+1];
    if (fread(filebuffer,1,l,f)!=(unsigned int)l) goto invalid;
    filebuffer[l]=0;
    if (strlen(filebuffer)<(unsigned int)l) goto invalid;
    if (strstr(filebuffer,"\\START92\\\r\n")) goto token;
  }
  invalid:
    fclose(f);
    puts("Invalid file.");
    return 3;  

  detoken: {
    fclose(f);
    if (OleInitialize(0)!=S_OK) {puts("OCX error. (#0)");return 4;}
    LCID lcid=GetUserDefaultLCID();
    GUID guid={0x07919935,0xC167,0x4F31,{0x86,0x9A,0x96,0xA9,0xF2,0x94,0x74,
                                         0x12}};
    ITypeLib *tlb;
    if (LoadRegTypeLib(&guid,1,0,lcid,&tlb)!=S_OK)
      {OleUninitialize();puts("OCX error. (#1)");return 4;}
    ITypeInfo *tinfo;
    MEMBERID mid;
    unsigned short n=1;
    if (tlb->lpVtbl->FindName(tlb,L"Tokens89OCX",0L,&tinfo,&mid,&n)!=S_OK || !n)
      {OleUninitialize();puts("OCX error. (#2)");return 4;}
    IDispatch *tokens89ocx;
    if (tinfo->lpVtbl->CreateInstance(tinfo,0,&IID_IDispatch,
        (void**)&tokens89ocx)!=S_OK)
      {OleUninitialize();puts("OCX error (#3).");return 4;}
    DISPID idOpenTI, idGetTIFile, idCROnlyToCRLF, idSaveTIGLAscii;
    if (tokens89ocx->lpVtbl->GetIDsOfNames(tokens89ocx,&IID_NULL,
                                           (OLECHAR *[1]){L"OpenTI"},1,
                                           lcid,&idOpenTI)!=S_OK
        || tokens89ocx->lpVtbl->GetIDsOfNames(tokens89ocx,&IID_NULL,
                                              (OLECHAR *[1]){L"GetTIFile"},1,
                                              lcid,&idGetTIFile)!=S_OK
        || tokens89ocx->lpVtbl->GetIDsOfNames(tokens89ocx,&IID_NULL,
                                              (OLECHAR *[1]){L"CROnlyToCRLF"},
                                              1,lcid,&idCROnlyToCRLF)!=S_OK
        || tokens89ocx->lpVtbl->GetIDsOfNames(tokens89ocx,&IID_NULL,
                                              (OLECHAR *[1]){L"SaveTIGLAscii"},
                                              1,lcid,&idSaveTIGLAscii)!=S_OK) {
        tokens89ocx->lpVtbl->Release(tokens89ocx);
        OleUninitialize();
        puts("OCX error (#4).");
        return 4;
      }
    int infilesize=strlen(argv[1])+1;
    int infilesizew=MultiByteToWideChar(CP_ACP,0,argv[1],infilesize,0,0);
    BSTR infile;
    {
      WCHAR infilew[infilesizew];
      MultiByteToWideChar(CP_ACP,0,argv[1],infilesize,infilew,infilesizew);
      infile=SysAllocString(infilew);
      if (!infile) {
        tokens89ocx->lpVtbl->Release(tokens89ocx);
        OleUninitialize();
        puts("OCX error (#5).");
        return 4;
      }
    }
    int outfilesize=strlen(argv[2])+1;
    int outfilesizew=MultiByteToWideChar(CP_ACP,0,argv[2],outfilesize,0,0);
    BSTR outfile;
    {
      WCHAR outfilew[outfilesizew];
      MultiByteToWideChar(CP_ACP,0,argv[2],outfilesize,outfilew,outfilesizew);
      outfile=SysAllocString(outfilew);
      if (!outfile) {
        SysFreeString(infile);
        tokens89ocx->lpVtbl->Release(tokens89ocx);
        OleUninitialize();
        puts("OCX error (#6).");
        return 4;
      }
    }
    VARIANT inf={{{VT_BSTR,0,0,0,{.bstrVal=infile}}}};
    VARIANT outf={{{VT_BSTR,0,0,0,{.bstrVal=outfile}}}};
    VARIANT data1,data2,name,dummy;
    if (tokens89ocx->lpVtbl->Invoke(tokens89ocx,idOpenTI,&IID_NULL,lcid,
                                    DISPATCH_METHOD,&(DISPPARAMS){&inf,0,1,0},
                                    &data1,0,0)!=S_OK) {
      SysFreeString(infile);
      SysFreeString(outfile);
      tokens89ocx->lpVtbl->Release(tokens89ocx);
      puts("Error reading file contents.");
      return 5;
    }
    if (tokens89ocx->lpVtbl->Invoke(tokens89ocx,idGetTIFile,&IID_NULL,lcid,
                                    DISPATCH_METHOD,&(DISPPARAMS){&inf,0,1,0},
                                    &name,0,0)!=S_OK) {
      SysFreeString(infile);
      SysFreeString(outfile);
      SysFreeString(data1.bstrVal);
      tokens89ocx->lpVtbl->Release(tokens89ocx);
      puts("Error reading file name.");
      return 6;
    }
    if (tokens89ocx->lpVtbl->Invoke(tokens89ocx,idCROnlyToCRLF,&IID_NULL,lcid,
                                    DISPATCH_METHOD,&(DISPPARAMS){&data1,0,1,0},
                                    &data2,0,0)!=S_OK) {
      SysFreeString(infile);
      SysFreeString(outfile);
      SysFreeString(data1.bstrVal);
      SysFreeString(name.bstrVal);
      tokens89ocx->lpVtbl->Release(tokens89ocx);
      puts("Error converting line endings.");
      return 7;
    }
    SysFreeString(data1.bstrVal);
    if (tokens89ocx->lpVtbl->Invoke(tokens89ocx,idSaveTIGLAscii,&IID_NULL,lcid,
                                    DISPATCH_METHOD,&(DISPPARAMS){(VARIANT [3])
                                    {data2,name,outf},0,3,0},&dummy,0,0)!=S_OK){
      SysFreeString(infile);
      SysFreeString(outfile);
      SysFreeString(data2.bstrVal);
      SysFreeString(name.bstrVal);
      tokens89ocx->lpVtbl->Release(tokens89ocx);
      puts("Error writing file.");
      return 8;
    }
    SysFreeString(infile);
    SysFreeString(outfile);
    SysFreeString(data2.bstrVal);
    SysFreeString(name.bstrVal);
    tokens89ocx->lpVtbl->Release(tokens89ocx);
    return 0;
  }

  token: {
    fclose(f);
    if (OleInitialize(0)!=S_OK) {puts("OCX error. (#0)");return 4;}
    LCID lcid=GetUserDefaultLCID();
    GUID guid={0x07919935,0xC167,0x4F31,{0x86,0x9A,0x96,0xA9,0xF2,0x94,0x74,
                                         0x12}};
    ITypeLib *tlb;
    if (LoadRegTypeLib(&guid,1,0,lcid,&tlb)!=S_OK)
      {OleUninitialize();puts("OCX error. (#7)");return 4;}
    ITypeInfo *tinfo;
    MEMBERID mid;
    unsigned short n=1;
    if (tlb->lpVtbl->FindName(tlb,L"Tokens89OCX",0L,&tinfo,&mid,&n)!=S_OK || !n)
      {OleUninitialize();puts("OCX error. (#8)");return 4;}
    IDispatch *tokens89ocx;
    if (tinfo->lpVtbl->CreateInstance(tinfo,0,&IID_IDispatch,
        (void**)&tokens89ocx)!=S_OK)
      {OleUninitialize();puts("OCX error (#9).");return 4;}
    DISPID idOpenTIGLAscii, idGetTIGLAsciiName, idCRLFToCROnly, idSaveTI;
    if (tokens89ocx->lpVtbl->GetIDsOfNames(tokens89ocx,&IID_NULL,
                                           (OLECHAR *[1]){L"OpenTIGLAscii"},1,
                                           lcid,&idOpenTIGLAscii)!=S_OK
        || tokens89ocx->lpVtbl->GetIDsOfNames(tokens89ocx,&IID_NULL,
                                              (OLECHAR *[1])
                                              {L"GetTIGLAsciiName"},1,lcid,
                                              &idGetTIGLAsciiName)!=S_OK
        || tokens89ocx->lpVtbl->GetIDsOfNames(tokens89ocx,&IID_NULL,
                                              (OLECHAR *[1]){L"CRLFToCROnly"},
                                              1,lcid,&idCRLFToCROnly)!=S_OK
        || tokens89ocx->lpVtbl->GetIDsOfNames(tokens89ocx,&IID_NULL,
                                              (OLECHAR *[1]){L"SaveTI"},1,lcid,
                                              &idSaveTI)!=S_OK) {
        tokens89ocx->lpVtbl->Release(tokens89ocx);
        OleUninitialize();
        puts("OCX error (#10).");
        return 4;
      }
    int infilesize=strlen(argv[1])+1;
    int infilesizew=MultiByteToWideChar(CP_ACP,0,argv[1],infilesize,0,0);
    BSTR infile;
    {
      WCHAR infilew[infilesizew];
      MultiByteToWideChar(CP_ACP,0,argv[1],infilesize,infilew,infilesizew);
      infile=SysAllocString(infilew);
      if (!infile) {
        tokens89ocx->lpVtbl->Release(tokens89ocx);
        OleUninitialize();
        puts("OCX error (#11).");
        return 4;
      }
    }
    int outfilesize=strlen(argv[2])+1;
    int outfilesizew=MultiByteToWideChar(CP_ACP,0,argv[2],outfilesize,0,0);
    BSTR outfile;
    {
      WCHAR outfilew[outfilesizew];
      MultiByteToWideChar(CP_ACP,0,argv[2],outfilesize,outfilew,outfilesizew);
      outfile=SysAllocString(outfilew);
      if (!outfile) {
        SysFreeString(infile);
        tokens89ocx->lpVtbl->Release(tokens89ocx);
        OleUninitialize();
        puts("OCX error (#12).");
        return 4;
      }
    }
    VARIANT inf={{{VT_BSTR,0,0,0,{.bstrVal=infile}}}};
    VARIANT outf={{{VT_BSTR,0,0,0,{.bstrVal=outfile}}}};
    VARIANT data1,data2,name,dummy;
    if (tokens89ocx->lpVtbl->Invoke(tokens89ocx,idOpenTIGLAscii,&IID_NULL,lcid,
                                    DISPATCH_METHOD,&(DISPPARAMS){&inf,0,1,0},
                                    &data1,0,0)!=S_OK) {
      SysFreeString(infile);
      SysFreeString(outfile);
      tokens89ocx->lpVtbl->Release(tokens89ocx);
      puts("Error reading file contents.");
      return 5;
    }
    if (tokens89ocx->lpVtbl->Invoke(tokens89ocx,idGetTIGLAsciiName,&IID_NULL,
                                    lcid,DISPATCH_METHOD,&(DISPPARAMS)
                                    {&inf,0,1,0},&name,0,0)!=S_OK) {
      SysFreeString(infile);
      SysFreeString(outfile);
      SysFreeString(data1.bstrVal);
      tokens89ocx->lpVtbl->Release(tokens89ocx);
      puts("Error reading file name.");
      return 6;
    }
    if (tokens89ocx->lpVtbl->Invoke(tokens89ocx,idCRLFToCROnly,&IID_NULL,lcid,
                                    DISPATCH_METHOD,&(DISPPARAMS){&data1,0,1,0},
                                    &data2,0,0)!=S_OK) {
      SysFreeString(infile);
      SysFreeString(outfile);
      SysFreeString(data1.bstrVal);
      SysFreeString(name.bstrVal);
      tokens89ocx->lpVtbl->Release(tokens89ocx);
      puts("Error converting line endings.");
      return 7;
    }
    SysFreeString(data1.bstrVal);
    BSTR folderstr=SysAllocString(L"main");
    if (!folderstr) {
      SysFreeString(infile);
      SysFreeString(outfile);
      SysFreeString(data2.bstrVal);
      SysFreeString(name.bstrVal);
      tokens89ocx->lpVtbl->Release(tokens89ocx);
      OleUninitialize();
      puts("Error writing file.");
      return 8;
    }
    VARIANT folder={{{VT_BSTR,0,0,0,{.bstrVal=folderstr}}}};
    VARIANT ftype={{{VT_I4,0,0,0,{.lVal=1}}}};
    if (tokens89ocx->lpVtbl->Invoke(tokens89ocx,idSaveTI,&IID_NULL,lcid,
                                    DISPATCH_METHOD,&(DISPPARAMS){(VARIANT [5])
                                    {ftype,data2,name,folder,outf},0,5,0},
                                    &dummy,0,0)!=S_OK){
      SysFreeString(infile);
      SysFreeString(outfile);
      SysFreeString(folderstr);
      SysFreeString(data2.bstrVal);
      SysFreeString(name.bstrVal);
      tokens89ocx->lpVtbl->Release(tokens89ocx);
      puts("Error writing file.");
      return 8;
    }
    SysFreeString(infile);
    SysFreeString(outfile);
    SysFreeString(folderstr);
    SysFreeString(data2.bstrVal);
    SysFreeString(name.bstrVal);
    tokens89ocx->lpVtbl->Release(tokens89ocx);
    return 0;
  }
}
